@@ -80,7 +80,7 @@ class BOTAN_UNSTABLE_API Client_Hello : public Handshake_Message {
       Client_Hello(Client_Hello&&) noexcept;
       Client_Hello& operator=(Client_Hello&&) noexcept;
 
-      ~Client_Hello();
+      ~Client_Hello() override;
 
       Handshake_Type type() const override;
 
@@ -109,6 +109,7 @@ class BOTAN_UNSTABLE_API Client_Hello : public Handshake_Message {
 
       std::vector<Group_Params> supported_ecc_curves() const;
 
+      // This returns any groups in the FFDHE range
       std::vector<Group_Params> supported_dh_groups() const;
 
       std::vector<Protocol_Version> supported_versions() const;
@@ -140,19 +141,19 @@ class BOTAN_UNSTABLE_API Client_Hello : public Handshake_Message {
       const std::vector<uint8_t>& compression_methods() const;
 
    protected:
-      std::unique_ptr<Client_Hello_Internal> m_data;
+      std::unique_ptr<Client_Hello_Internal> m_data;  // NOLINT(*non-private-member-variable*)
 };
 
 class BOTAN_UNSTABLE_API Client_Hello_12 final : public Client_Hello {
    public:
       class Settings final {
          public:
-            Settings(const Protocol_Version version, std::string_view hostname = "") :
+            explicit Settings(const Protocol_Version version, std::string_view hostname = "") :
                   m_new_session_version(version), m_hostname(hostname) {}
 
-            const Protocol_Version protocol_version() const { return m_new_session_version; }
+            Protocol_Version protocol_version() const { return m_new_session_version; }
 
-            const std::string hostname() const { return m_hostname; }
+            const std::string& hostname() const { return m_hostname; }
 
          private:
             const Protocol_Version m_new_session_version;
@@ -182,7 +183,7 @@ class BOTAN_UNSTABLE_API Client_Hello_12 final : public Client_Hello {
 
    protected:
       friend class Client_Hello_13;  // to allow construction by Client_Hello_13::parse()
-      Client_Hello_12(std::unique_ptr<Client_Hello_Internal> data);
+      explicit Client_Hello_12(std::unique_ptr<Client_Hello_Internal> data);
 
    public:
       using Client_Hello::compression_methods;
@@ -207,6 +208,9 @@ class BOTAN_UNSTABLE_API Client_Hello_12 final : public Client_Hello {
       bool supports_encrypt_then_mac() const;
 
       void update_hello_cookie(const Hello_Verify_Request& hello_verify);
+
+   private:
+      void add_tls12_supported_groups_extensions(const Policy& policy);
 };
 
 #if defined(BOTAN_HAS_TLS_13)
@@ -223,7 +227,8 @@ class BOTAN_UNSTABLE_API Client_Hello_13 final : public Client_Hello {
                       RandomNumberGenerator& rng,
                       std::string_view hostname,
                       const std::vector<std::string>& next_protocols,
-                      std::optional<Session_with_Handle>& session);
+                      std::optional<Session_with_Handle>& session,
+                      std::vector<ExternalPSK> psks);
 
       static std::variant<Client_Hello_13, Client_Hello_12> parse(const std::vector<uint8_t>& buf);
 
@@ -234,7 +239,7 @@ class BOTAN_UNSTABLE_API Client_Hello_13 final : public Client_Hello {
 
       /**
        * Select the highest protocol version from the list of versions
-       * supported by the client. If no such version can be determind this
+       * supported by the client. If no such version can be determined this
        * returns std::nullopt.
        */
       std::optional<Protocol_Version> highest_supported_version(const Policy& policy) const;
@@ -247,7 +252,7 @@ class BOTAN_UNSTABLE_API Client_Hello_13 final : public Client_Hello {
       void validate_updates(const Client_Hello_13& new_ch);
 
    private:
-      Client_Hello_13(std::unique_ptr<Client_Hello_Internal> data);
+      explicit Client_Hello_13(std::unique_ptr<Client_Hello_Internal> data);
 
       /**
        * If the Client Hello contains a PSK extensions with identities this will
@@ -273,7 +278,7 @@ class BOTAN_UNSTABLE_API Server_Hello : public Handshake_Message {
       Server_Hello(Server_Hello&&) noexcept;
       Server_Hello& operator=(Server_Hello&&) noexcept;
 
-      ~Server_Hello();
+      ~Server_Hello() override;
 
       std::vector<uint8_t> serialize() const override;
 
@@ -296,7 +301,7 @@ class BOTAN_UNSTABLE_API Server_Hello : public Handshake_Message {
       Protocol_Version legacy_version() const;
 
    protected:
-      std::unique_ptr<Server_Hello_Internal> m_data;
+      std::unique_ptr<Server_Hello_Internal> m_data;  // NOLINT(*non-private-member-variable*)
 };
 
 class BOTAN_UNSTABLE_API Server_Hello_12 final : public Server_Hello {
@@ -416,6 +421,7 @@ class BOTAN_UNSTABLE_API Server_Hello_13 : public Server_Hello {
       Server_Hello_13(const Client_Hello_13& ch,
                       std::optional<Named_Group> key_exchange_group,
                       Session_Manager& session_mgr,
+                      Credentials_Manager& credentials_mgr,
                       RandomNumberGenerator& rng,
                       Callbacks& cb,
                       const Policy& policy);
@@ -426,6 +432,7 @@ class BOTAN_UNSTABLE_API Server_Hello_13 : public Server_Hello {
       static std::variant<Hello_Retry_Request, Server_Hello_13> create(const Client_Hello_13& ch,
                                                                        bool hello_retry_request_allowed,
                                                                        Session_Manager& session_mgr,
+                                                                       Credentials_Manager& credentials_mgr,
                                                                        RandomNumberGenerator& rng,
                                                                        const Policy& policy,
                                                                        Callbacks& cb);
@@ -440,7 +447,7 @@ class BOTAN_UNSTABLE_API Server_Hello_13 : public Server_Hello {
       /**
        * @returns the selected version as indicated by the supported_versions extension
        */
-      Protocol_Version selected_version() const override final;
+      Protocol_Version selected_version() const final;
 };
 
 class BOTAN_UNSTABLE_API Hello_Retry_Request final : public Server_Hello_13 {
@@ -481,6 +488,11 @@ class BOTAN_UNSTABLE_API Client_Key_Exchange final : public Handshake_Message {
 
       const secure_vector<uint8_t>& pre_master_secret() const { return m_pre_master; }
 
+      /**
+       * @returns the agreed upon PSK identity or std::nullopt if not applicable
+       */
+      const std::optional<std::string>& psk_identity() const { return m_psk_identity; }
+
       Client_Key_Exchange(Handshake_IO& io,
                           Handshake_State& state,
                           const Policy& policy,
@@ -501,6 +513,7 @@ class BOTAN_UNSTABLE_API Client_Key_Exchange final : public Handshake_Message {
 
       std::vector<uint8_t> m_key_material;
       secure_vector<uint8_t> m_pre_master;
+      std::optional<std::string> m_psk_identity;
 };
 
 /**
@@ -535,12 +548,27 @@ class Certificate_Request_13;
 */
 class BOTAN_UNSTABLE_API Certificate_13 final : public Handshake_Message {
    public:
-      struct Certificate_Entry {
-            // TODO: RFC 8446 4.4.2 specifies the possibility to negotiate the usage
-            //       of a single raw public key in lieu of the X.509 certificate
-            //       chain. This is left for future work.
-            X509_Certificate certificate;
-            Extensions extensions;
+      class Certificate_Entry {
+         public:
+            Certificate_Entry(TLS_Data_Reader& reader, Connection_Side side, Certificate_Type cert_type);
+            explicit Certificate_Entry(X509_Certificate cert);
+            explicit Certificate_Entry(std::shared_ptr<Public_Key> raw_public_key);
+
+            bool has_certificate() const { return m_certificate.has_value(); }
+
+            const X509_Certificate& certificate() const;
+            std::shared_ptr<const Public_Key> public_key() const;
+
+            std::vector<uint8_t> serialize() const;
+
+            Extensions& extensions() { return m_extensions; }
+
+            const Extensions& extensions() const { return m_extensions; }
+
+         private:
+            std::optional<X509_Certificate> m_certificate;
+            std::shared_ptr<Public_Key> m_raw_public_key;
+            Extensions m_extensions;
       };
 
    public:
@@ -548,10 +576,14 @@ class BOTAN_UNSTABLE_API Certificate_13 final : public Handshake_Message {
 
       std::vector<X509_Certificate> cert_chain() const;
 
+      bool has_certificate_chain() const;
+      bool is_raw_public_key() const;
+
       size_t count() const { return m_entries.size(); }
 
       bool empty() const { return m_entries.empty(); }
 
+      std::shared_ptr<const Public_Key> public_key() const;
       const X509_Certificate& leaf() const;
 
       const std::vector<uint8_t>& request_context() const { return m_request_context; }
@@ -563,7 +595,8 @@ class BOTAN_UNSTABLE_API Certificate_13 final : public Handshake_Message {
       Certificate_13(const Certificate_Request_13& cert_request,
                      std::string_view hostname,
                      Credentials_Manager& credentials_manager,
-                     Callbacks& callbacks);
+                     Callbacks& callbacks,
+                     Certificate_Type cert_type);
 
       /**
        * Create a Server Certificate message
@@ -572,22 +605,28 @@ class BOTAN_UNSTABLE_API Certificate_13 final : public Handshake_Message {
        */
       Certificate_13(const Client_Hello_13& client_hello,
                      Credentials_Manager& credentials_manager,
-                     Callbacks& callbacks);
+                     Callbacks& callbacks,
+                     Certificate_Type cert_type);
 
       /**
       * Deserialize a Certificate message
       * @param buf the serialized message
       * @param policy the TLS policy
       * @param side is this a Connection_Side::Server or Connection_Side::Client certificate message
+      * @param cert_type is the certificate type that was negotiated during the handshake
       */
-      Certificate_13(const std::vector<uint8_t>& buf, const Policy& policy, const Connection_Side side);
+      Certificate_13(const std::vector<uint8_t>& buf,
+                     const Policy& policy,
+                     Connection_Side side,
+                     Certificate_Type cert_type);
 
       /**
       * Validate a Certificate message regarding what extensions are expected based on
-      * previous handshake messages. Also call the tls_examine_extenions() callback
+      * previous handshake messages. Also call the tls_examine_extensions() callback
       * for each entry.
       *
       * @param requested_extensions Extensions of Client_Hello or Certificate_Request messages
+      * @param cb Callback that will be called for each extension.
       */
       void validate_extensions(const std::set<Extension_Code>& requested_extensions, Callbacks& cb) const;
 
@@ -608,6 +647,14 @@ class BOTAN_UNSTABLE_API Certificate_13 final : public Handshake_Message {
       void setup_entries(std::vector<X509_Certificate> cert_chain,
                          const Certificate_Status_Request* csr,
                          Callbacks& callbacks);
+      void setup_entry(std::shared_ptr<Public_Key> raw_public_key, Callbacks& callbacks);
+
+      void verify_certificate_chain(Callbacks& callbacks,
+                                    const Policy& policy,
+                                    Credentials_Manager& creds,
+                                    std::string_view hostname,
+                                    bool use_ocsp,
+                                    Usage_Type usage_type) const;
 
    private:
       std::vector<uint8_t> m_request_context;
@@ -628,7 +675,7 @@ class BOTAN_UNSTABLE_API Certificate_Status final : public Handshake_Message {
 
       const std::vector<uint8_t>& response() const { return m_response; }
 
-      explicit Certificate_Status(const std::vector<uint8_t>& buf, const Connection_Side from);
+      explicit Certificate_Status(const std::vector<uint8_t>& buf, Connection_Side from);
 
       Certificate_Status(Handshake_IO& io, Handshake_Hash& hash, const OCSP::Response& response);
 
@@ -637,7 +684,7 @@ class BOTAN_UNSTABLE_API Certificate_Status final : public Handshake_Message {
        */
       Certificate_Status(Handshake_IO& io, Handshake_Hash& hash, std::vector<uint8_t> raw_response_bytes);
 
-      Certificate_Status(std::vector<uint8_t> raw_response_bytes);
+      explicit Certificate_Status(std::vector<uint8_t> raw_response_bytes);
 
       std::vector<uint8_t> serialize() const override;
 
@@ -679,7 +726,7 @@ class BOTAN_UNSTABLE_API Certificate_Request_13 final : public Handshake_Message
    public:
       Handshake_Type type() const override;
 
-      Certificate_Request_13(const std::vector<uint8_t>& buf, const Connection_Side side);
+      Certificate_Request_13(const std::vector<uint8_t>& buf, Connection_Side side);
 
       //! Creates a Certificate_Request message if it is required by the configuration
       //! @return std::nullopt if configuration does not require client authentication
@@ -696,7 +743,7 @@ class BOTAN_UNSTABLE_API Certificate_Request_13 final : public Handshake_Message
 
       std::vector<uint8_t> serialize() const override;
 
-      const std::vector<uint8_t> context() const { return m_context; }
+      const std::vector<uint8_t>& context() const { return m_context; }
 
    private:
       Certificate_Request_13(std::vector<X509_DN> acceptable_CAs, const Policy& policy, Callbacks& callbacks);
@@ -714,14 +761,14 @@ class BOTAN_UNSTABLE_API Certificate_Verify : public Handshake_Message {
 
       Signature_Scheme signature_scheme() const { return m_scheme; }
 
-      Certificate_Verify(const std::vector<uint8_t>& buf);
+      explicit Certificate_Verify(const std::vector<uint8_t>& buf);
       Certificate_Verify() = default;
 
       std::vector<uint8_t> serialize() const override;
 
    protected:
-      std::vector<uint8_t> m_signature;
-      Signature_Scheme m_scheme;
+      std::vector<uint8_t> m_signature;  // NOLINT(*non-private-member-variable*)
+      Signature_Scheme m_scheme;         // NOLINT(*non-private-member-variable*)
 };
 
 /**
@@ -758,7 +805,7 @@ class BOTAN_UNSTABLE_API Certificate_Verify_13 final : public Certificate_Verify
       * @param buf the serialized message
       * @param side is this a Connection_Side::Server or Connection_Side::Client certificate message
       */
-      Certificate_Verify_13(const std::vector<uint8_t>& buf, const Connection_Side side);
+      Certificate_Verify_13(const std::vector<uint8_t>& buf, Connection_Side side);
 
       Certificate_Verify_13(const Certificate_13& certificate_message,
                             const std::vector<Signature_Scheme>& peer_allowed_schemes,
@@ -770,7 +817,7 @@ class BOTAN_UNSTABLE_API Certificate_Verify_13 final : public Certificate_Verify
                             Callbacks& callbacks,
                             RandomNumberGenerator& rng);
 
-      bool verify(const X509_Certificate& cert, Callbacks& callbacks, const Transcript_Hash& transcript_hash) const;
+      bool verify(const Public_Key& public_key, Callbacks& callbacks, const Transcript_Hash& transcript_hash) const;
 
    private:
       Connection_Side m_side;
@@ -793,7 +840,7 @@ class BOTAN_UNSTABLE_API Finished : public Handshake_Message {
 
    protected:
       using Handshake_Message::Handshake_Message;
-      std::vector<uint8_t> m_verification_data;
+      std::vector<uint8_t> m_verification_data;  // NOLINT(*non-private-member-variable*)
 };
 
 class BOTAN_UNSTABLE_API Finished_12 final : public Finished {
@@ -909,7 +956,7 @@ class BOTAN_UNSTABLE_API New_Session_Ticket_12 final : public Handshake_Message
       std::vector<uint8_t> serialize() const override;
 
    private:
-      std::chrono::seconds m_ticket_lifetime_hint;
+      std::chrono::seconds m_ticket_lifetime_hint{};
       Session_Ticket m_ticket;
 };
 
@@ -955,7 +1002,7 @@ class BOTAN_UNSTABLE_API New_Session_Ticket_13 final : public Handshake_Message
       //    of time than what is stated in the ticket_lifetime.
       //
       // ... hence we call it 'lifetime hint'.
-      std::chrono::seconds m_ticket_lifetime_hint;
+      std::chrono::seconds m_ticket_lifetime_hint{};
       uint32_t m_ticket_age_add;
       Ticket_Nonce m_ticket_nonce;
       Opaque_Session_Handle m_handle;
@@ -980,7 +1027,7 @@ class BOTAN_UNSTABLE_API Key_Update final : public Handshake_Message {
    public:
       Handshake_Type type() const override { return Handshake_Type::KeyUpdate; }
 
-      explicit Key_Update(const bool request_peer_update);
+      explicit Key_Update(bool request_peer_update);
       explicit Key_Update(const std::vector<uint8_t>& buf);
 
       std::vector<uint8_t> serialize() const override;
@@ -991,7 +1038,7 @@ class BOTAN_UNSTABLE_API Key_Update final : public Handshake_Message {
       bool m_update_requested;
 };
 
-namespace {
+namespace detail {
 template <typename T>
 struct as_wrapped_references {};
 
@@ -1002,7 +1049,7 @@ struct as_wrapped_references<std::variant<AlternativeTs...>> {
 
 template <typename T>
 using as_wrapped_references_t = typename as_wrapped_references<T>::type;
-}  // namespace
+}  // namespace detail
 
 // Handshake message types from RFC 8446 4.
 using Handshake_Message_13 = std::variant<Client_Hello_13,
@@ -1016,7 +1063,7 @@ using Handshake_Message_13 = std::variant<Client_Hello_13,
                                           Certificate_Request_13,
                                           Certificate_Verify_13,
                                           Finished_13>;
-using Handshake_Message_13_Ref = as_wrapped_references_t<Handshake_Message_13>;
+using Handshake_Message_13_Ref = detail::as_wrapped_references_t<Handshake_Message_13>;
 
 using Post_Handshake_Message_13 = std::variant<New_Session_Ticket_13, Key_Update>;
 
@@ -1034,14 +1081,14 @@ using Server_Handshake_13_Message = std::variant<Server_Hello_13,
                                                  Certificate_Request_13,
                                                  Certificate_Verify_13,
                                                  Finished_13>;
-using Server_Handshake_13_Message_Ref = as_wrapped_references_t<Server_Handshake_13_Message>;
+using Server_Handshake_13_Message_Ref = detail::as_wrapped_references_t<Server_Handshake_13_Message>;
 
 using Client_Handshake_13_Message = std::variant<Client_Hello_13,
                                                  Client_Hello_12,  // indicates a TLS peer that does not offer TLS 1.3
                                                  Certificate_13,
                                                  Certificate_Verify_13,
                                                  Finished_13>;
-using Client_Handshake_13_Message_Ref = as_wrapped_references_t<Client_Handshake_13_Message>;
+using Client_Handshake_13_Message_Ref = detail::as_wrapped_references_t<Client_Handshake_13_Message>;
 
 #endif  // BOTAN_HAS_TLS_13
 
